#include "mpi.h"
#include <stdio.h>

#define MAX_COUNT 1000

int main( argc, argv )
int argc;
char **argv;
{
  MPI_File     fh;
  MPI_Info     info;
  MPI_Status   status;
  MPI_Datatype filetype, t1;
  MPI_Comm     cartcomm;
  char         *filename;
  int          buffer[MAX_COUNT];
  int          count, rank;
  int          ndims[2];

  MPI_Init( &argc, &argv );
  
  MPI_Info_create( &info );
  /* MPI_INFO_NULL may be used instead of this info */
  MPI_Info_set( info, "collective_buffering", "true" );
  MPI_Info_set( info, "access_style", "write_once" );
  MPI_Info_set( info, "file_perm", "0666" );

  /* The filename is implementation-specific (just as it is
     for all environments).  This assumes that a simple file name will 
     work.  The file name may be over-ridden from the command line */
  filename = "sample.input";
  if (argc > 1 && argv[1]) filename = argv[1];

  /* 
     Here's the hard part - forming our decomposition of a 2-d array across
     all of the processors.  Fortunately, MPI-2 provides functions to make
     this a little easier.

   */
  ndims[0] = ndims[1] = 0;
  MPI_Dims_create( , ndims );
  /* We build a ndims[0] x ndims[1] decomposition */
  MPI_Cart_create( ..., ndims, &cartcomm );
  MPI_Cart_coords( ..., cartcomm, coords );

  /* Sizes of global array in each dimension */
  sizes[0] = 100;
  sizes[1] = 100;
  /* Size of local array in each dimension */
  subsizes[0] = sizes[0] / ndims[0];
  subsizes[1] = sizes[1] / ndims[1];
  /* Since the sizes may not divide exactly, we update these to handle the
     leftovers */
  if (coords[0] < sizes[0] % ndims[0]) subsizes[0]++;
  if (coords[1] < sizes[1] % ndims[1]) subsizes[1]++;

  /* Index, relative to global array, of the initial block */
  starts[0] = coords[0] * sizes[0] / ndims[0];
  starts[1] = coords[1] * sizes[1] / ndims[1];
  /* If the sizes divide exactly, we can just multiply the coords
     by the size as above.  If not, we need to know the sizes of the 
     blocks before us.  This can be done locally by using the same size
     rule used above.  An alternative is to use MPI_Exscan, applied to a 
     communicator formed for each row and column of the decomposition.  
     This will be more expensive but may be clearer.
   */
  
  nudge = sizes[0] % ndims[0];
  if (coords[0] < nudge) nudge = coords[0] + 1;
  starts[0] += nudge;
  nudge = sizes[1] % ndims[1];
  if (coords[1] < nudge) nudge = coords[1] + 1;
  starts[1] += nudge;
  MPI_Type_subarray_create( 2, sizes, subsizes, starts, MPI_ORDER_C, 
			    MPI_INT, &filetype );
  MPI_Type_commit( &filetype );
  
  /* All processes open the file.  */
  MPI_File_open( MPI_COMM_WORLD, filename, 
		 MPI_MODE_WRONLY | MPI_MODE_UNIQUE_OPEN | MPI_MODE_CREATE,
		 info, &fh );

  /* Give each process a different part of the data */
  --- assume that the displacement (0 below) is handled by the datatype
      above ---
  MPI_File_set_view( fh, 0, MPI_INT, filetype, "native", MPI_INFO_NULL );

  /* Generate data to write.  This numbers the local buffer with the global
     indices */
  r = subsizes[0];
  c = subsizes[1];
  for (i=0; i < r ; i++) {
    for (j=0; j < c; j++) {
      buffer[i* + j] = (i + starts[]) * sizes[0] + (j + starts[]); 
    }
  }
  /* Now, each processor can write it's part of the array */
  MPI_File_write_all( fh, buffer, MAX_COUNT, MPI_INT, &status );

  MPI_Comm_rank( MPI_COMM_WORLD, &rank );
  if (rank == 0) {
    printf( "Wrote array\n", count );
  }

  MPI_File_close( &fh );
  MPI_Type_free( &filetype );

  MPI_Finalize( );
  return 0;
}
