#include <stdio.h>
#include <stdlib.h>
#include "mpi.h"

/* 
   This is the number of times each test is run; the time presented is the 
   minimum of the tests
 */
#define NUMBER_OF_TESTS 10

/* Forward declarations */
void   TestOp();
double TestAllreduce();
double TestBcast();
double TestScan();
double TestAlltoAll();
double TestScatter();
double TestGather();
double TestAllGather();

int main( argc, argv )
int argc;
char **argv;
{
    int          rank, size;
    int          j, k, nloop, max_elm;

    MPI_Init( &argc, &argv );

    MPI_Comm_rank( MPI_COMM_WORLD, &rank );
    MPI_Comm_size( MPI_COMM_WORLD, &size );
    if (rank == 0)
	printf( "Kind      \tn_elm\ttime (sec)\n" );

    max_elm = 130000;

    TestOp( TestBcast,     "Bcast         ", max_elm );
    TestOp( TestAllreduce, "Allreduce(sum)", max_elm );
    TestOp( TestScan,      "Scan(sum)     ", max_elm );

    /* These have unbalanced memory needs - the root process has size times
       as much data.  In this case, we make the ROOT size max_elm */
    TestOp( TestScatter,   "Scatter       ", max_elm / size );
    TestOp( TestGather,    "Gather        ", max_elm / size );

    /* These don't have a root, but ALL processes need size * max_elm
       for the delivered data */
#ifdef FOO
    TestOp( TestAllGather, "Allgather     ", max_elm / size );
    TestOp( TestAlltoAll,  "AllToAll      ", max_elm / size );
#endif
    /* There are additional routines, like MPI_Reduce_scatter and the 'v'
       routine (e.g., MPI_Scatterv) */

    MPI_Finalize( );
    return 0;
}

void MemAbort( routine, size )
char *routine;
int  size;
{
  fprintf( stderr, "Could not allocate %d doubles\n", size );
  MPI_Abort( 1, MPI_COMM_WORLD );
  exit( 1 );
}

void TestOp( routine, name, max_elm )
double (*routine)();
char    *name;
int     max_elm;
{
  int    nloop = 1000;
  int    n_elm, rank;
  double tmin;

  MPI_Comm_rank( MPI_COMM_WORLD, &rank );

  /* For large n_elm, we  want to reduce nloop.  We need a deterministic
     method since all processes need the same value of nloop */
  n_elm = 1;
  while (n_elm < max_elm) {
    tmin = (*routine)( n_elm, nloop, MPI_COMM_WORLD );

    if (rank == 0) {
	printf( "%s\t%d\t%f\n", name, n_elm, tmin );
    }
    n_elm *= 2;
    if (nloop > 16 && n_elm > 100) nloop = nloop / 2;
  }
}

/*
 * Individual Tests
 */
double TestAllreduce( n_elm, nloop, comm )
int      n_elm, nloop;
MPI_Comm comm;
{
  double tmin, t1, t2;
  int k, j;
  double *bufin, *bufout;
  
  bufin = (double *) malloc( n_elm * sizeof(double) );
  bufout = (double *) malloc( n_elm * sizeof(double) );
  if (!bufin || !bufout) MemAbort( "Allreduce", n_elm );
  for (k=0; k<n_elm; k++) {
    bufin[k] = 1.0;
    bufout[k] = -1.0;
  }

  tmin  = 1000;
  for (k=0; k<NUMBER_OF_TESTS; k++) {
    MPI_Barrier( comm );
    t1 = MPI_Wtime();
    for (j=0; j<nloop; j++) {
      MPI_Allreduce( bufin, bufout, n_elm, MPI_DOUBLE, MPI_SUM, comm );
    }
    t2 = (MPI_Wtime() - t1) / nloop;
    if (t2 < tmin) tmin = t2;
  }
  free( bufout );
  free( bufin );
  return tmin;
}

double TestScan( n_elm, nloop, comm )
int      n_elm, nloop;
MPI_Comm comm;
{
  double tmin, t1, t2;
  int k, j;
  double *bufin, *bufout;
  
  bufin = (double *) malloc( n_elm * sizeof(double) );
  bufout = (double *) malloc( n_elm * sizeof(double) );
  if (!bufin || !bufout) MemAbort( "Scan", n_elm );
  for (k=0; k<n_elm; k++) {
    bufin[k] = 1.0;
    bufout[k] = -1.0;
  }

  tmin  = 1000;
  for (k=0; k<NUMBER_OF_TESTS; k++) {
    MPI_Barrier( comm );
    t1 = MPI_Wtime();
    for (j=0; j<nloop; j++) {
      MPI_Scan( bufin, bufout, n_elm, MPI_DOUBLE, MPI_SUM, comm );
    }
    t2 = (MPI_Wtime() - t1) / nloop;
    if (t2 < tmin) tmin = t2;
  }
  free( bufout );
  free( bufin );
  return tmin;
}

double TestBcast( n_elm, nloop, comm )
int      n_elm, nloop;
MPI_Comm comm;
{
  double tmin, t1, t2;
  int k, j;
  double *bufin;
  
  bufin = (double *) malloc( n_elm * sizeof(double) );
  if (!bufin) MemAbort( "Bcast", n_elm );
  for (k=0; k<n_elm; k++) {
    bufin[k] = 1.0;
  }

  tmin  = 1000;
  for (k=0; k<NUMBER_OF_TESTS; k++) {
    MPI_Barrier( comm );
    t1 = MPI_Wtime();
    for (j=0; j<nloop; j++) {
      MPI_Bcast( bufin, n_elm, MPI_DOUBLE, 0, comm );
    }
    t2 = (MPI_Wtime() - t1) / nloop;
    if (t2 < tmin) tmin = t2;
  }
  free( bufin );
  return tmin;
}

double TestScatter( n_elm, nloop, comm )
int      n_elm, nloop;
MPI_Comm comm;
{
  double tmin, t1, t2;
  int k, j, size, rank;
  double *bufin, *bufout;
    
  MPI_Comm_size( comm, &size );
  MPI_Comm_rank( comm, &rank );

  bufin = (double *) malloc( n_elm * size * sizeof(double) );
  bufout = (double *) malloc( n_elm * sizeof(double) );
  if (!bufin || !bufout) MemAbort( "Scatter", n_elm );
  for (k=0; k<n_elm*size; k++) {
    bufin[k] = 1.0;
  }
  for (k=0; k<n_elm; k++) {
    bufout[k] = -1.0;
  }

  tmin  = 1000;
  for (k=0; k<NUMBER_OF_TESTS; k++) {
    MPI_Barrier( comm );
    t1 = MPI_Wtime();
    /* Interesting question is alternate initial rank */
    for (j=0; j<nloop; j++) {
      MPI_Scatter( bufin, n_elm, MPI_DOUBLE, 
		   bufout, n_elm, MPI_DOUBLE, 0, comm );
    }
    t2 = (MPI_Wtime() - t1) / nloop;
    if (t2 < tmin) tmin = t2;
  }
  free( bufout );
  free( bufin );
  return tmin;
}

double TestGather( n_elm, nloop, comm )
int      n_elm, nloop;
MPI_Comm comm;
{
  double tmin, t1, t2;
  int k, j, size, rank;
  double *bufin, *bufout;
  
  MPI_Comm_size( comm, &size );
  MPI_Comm_rank( comm, &rank );
  bufout = (double *) malloc( n_elm * size * sizeof(double) );
  bufin  = (double *) malloc( n_elm * sizeof(double) );
  if (!bufin || !bufout) MemAbort( "Gather", n_elm );
  for (k=0; k<n_elm*size; k++) {
    bufout[k] = 1.0;
  }
  for (k=0; k<n_elm; k++) {
    bufin[k] = -1.0;
  }

  tmin  = 1000;
  for (k=0; k<NUMBER_OF_TESTS; k++) {
    MPI_Barrier( comm );
    t1 = MPI_Wtime();
    /* Interesting question is alternate initial rank */
    for (j=0; j<nloop; j++) {
      MPI_Gather( bufin, n_elm, MPI_DOUBLE, 
		  bufout, n_elm, MPI_DOUBLE, 0, comm );
    }
    t2 = (MPI_Wtime() - t1) / nloop;
    if (t2 < tmin) tmin = t2;
  }
  free( bufout );
  free( bufin );
  return tmin;
}
